/// @addtogroup struct
/// @{

/////////////////////////////////////////////////////////////////////////////
///
/// @file bitsets.h
///
/// This file defines a class that uses bit representation of a set
/// to store many small sets.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2010 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started on February 22, 2007. Last revision: February 22, 2007.


#ifndef _CHOMP_STRUCT_BITSETS_H_
#define _CHOMP_STRUCT_BITSETS_H_

#include "chomp/system/config.h"

#include <iostream>
#include <cstdlib>

namespace chomp {
namespace homology {


// --------------------------------------------------
// ------------------- BitSets ---------------------
// --------------------------------------------------

/// This class uses bit representation to store many small sets.
/// Each of the sets can contain integer numbers from 0 to the limit
/// specified in the constructor. The number of sets must also be given
/// in the constructor. A contiguous chunk of memory is used to store
/// the bits that represent the sets.
/// This class is not fool-proof, so use it carefully.
/// Minimal interface, sorry.
class BitSets
{
public:
	/// Constructor of a family of empty sets.
	/// The number of the sets and the maximal number of elements
	/// in each set must be provided at initialization.
	BitSets (int _nsets, int _nelem);

	/// Copy constructor.
	BitSets (const BitSets &s);

	/// Assignment operator.
	BitSets &operator = (const BitSets &s);

	/// Destructor.
	~BitSets ();

	/// Adds an element to the given set.
	void add (int n, int e);

	/// Removes an element from the given set.
	void remove (int n, int e);

	/// Checks if an element belongs to the given set.
	bool check (int n, int e) const;

	/// Adds the entire set 'm' to the set 'n'.
	void addset (int n, int m);

	/// Retrieves an element >= 'start' in the given set.
	/// Returns -1 if none.
	int get (int n, int start = 0) const;

private:
	/// The number of sets.
	int nsets;
	
	/// The maximal number of elements in each set.
	int nelem;

	/// The number of bytes used for each set.
	int nbytes;

	/// The memory buffer for storing the bits.
	unsigned char *bits;

}; /* BitSets */

// --------------------------------------------------

inline BitSets::BitSets (int _nsets, int _nelem):
	nsets (_nsets), nelem (_nelem), nbytes ((_nelem + 7) >> 3), bits (0)
{
	int size = nsets * nbytes;
	bits = new unsigned char [size];
	for (int i = 0; i < size; ++ i)
		bits [i] = 0;
	return;
} /* BitSets::BitSets */

inline BitSets::BitSets (const BitSets &s):
	nsets (s. nsets), nelem (s. nelem), nbytes (s. nbytes), bits (0)
{
	int size = nsets * nbytes;
	bits = new unsigned char [size];
	for (int i = 0; i < size; ++ i)
		bits [i] = s. bits [i];
	return;
} /* BitSets::BitSets */

inline BitSets &BitSets::operator = (const BitSets &s)
{
	delete [] bits;
	nsets = s. nsets;
	nelem = s. nelem;
	nbytes = s. nbytes;
	int size = nsets * nbytes;
	bits = new unsigned char [size];
	for (int i = 0; i < size; ++ i)
		bits [i] = s. bits [i];
	return *this;
} /* BitSets::operator = */

inline BitSets::~BitSets ()
{
	delete [] bits;
	return;
} /* BitSets::~BitSets */

inline void BitSets::add (int n, int e)
{
//	sbug << "Add " << e << " to " << n << ".\n";
	unsigned char *buf = bits + n * nbytes;
	buf [e >> 3] |= (unsigned char) (0x01 << (e & 0x07));
	return;
} /* BitSets::add */

inline void BitSets::remove (int n, int e)
{
	unsigned char *buf = bits + n * nbytes;
	buf [e >> 3] &= (unsigned char) ~(0x01 << (e & 0x07));
	return;
} /* BitSets::remove */

inline bool BitSets::check (int n, int e) const
{
	unsigned char *buf = bits + n * nbytes;
	return !!(buf [e >> 3] & (0x01 << (e & 0x07)));
} /* BitSets::check */

inline void BitSets::addset (int n, int m)
{
	unsigned char *nbuf = bits + n * nbytes;
	unsigned char *mbuf = bits + m * nbytes;
	for (int i = 0; i < nbytes; ++ i)
		*(nbuf ++) |= *(mbuf ++);
	return;
} /* BitSets::addset */

inline int BitSets::get (int n, int start) const
{
	if (start >= nelem)
		return -1;
	unsigned char *buf = bits + n * nbytes;
	int offset = start >> 3;
	int bit = start & 0x07;
	if (buf [offset] & (0xFF << bit))
	{
		for (; bit < 8; ++ bit)
		{
			if (buf [offset] & (0x01 << bit))
				return (offset << 3) + bit;
		}
		throw "Wrong bit number in BitSets.";
	}
	for (++ offset; offset < nbytes; ++ offset)
	{
		if (!buf [offset])
			continue;
		for (int bit = 0; bit < 8; ++ bit)
		{
			if (buf [offset] & (0x01 << bit))
				return (offset << 3) + bit;
		}
		throw "False bit in BitSets.";
	}
	return -1;
} /* BitSets::get */


} // namespace homology
} // namespace chomp

#endif // _CHOMP_STRUCT_BITSETS_H_

/// @}

