/// @addtogroup struct
/// @{

/////////////////////////////////////////////////////////////////////////////
///
/// @file bitfield.cpp
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2010 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started in April 1999. Last revision: September 2, 2003.


#include "chomp/system/config.h"
#include "chomp/struct/bitfield.h"

#include <iostream>
#include <cstdlib>

namespace chomp {
namespace homology {


// --------------------------------------------------
// ------------------- BITFIELD ---------------------
// --------------------------------------------------

void BitField::define (unsigned char *buf, int length)
{
	table = buf;
	register int len = (length + 7) >> 3;
	for (int i = 0; i < len; i ++)
		table [i] = 0;
	return;
} /* BitField::define */

void BitField::allocate (int length)
{
	register int len = (length + 7) >> 3;
	table = new unsigned char [len];
	if (!table)
		throw "Not enough memory for a bit field.";
	for (int i = 0; i < len; i ++)
		table [i] = 0;
	return;
} /* BitField::allocate */

void BitField::takebits (const BitField &from, int length)
{
	register int len = (length + 7) >> 3;
	for (int i = 0; i < len; i ++)
		table [i] = from. table [i];
	return;
} /* BitField::takebits */

void BitField::clearall (int length)
{
	register int len = (length + 7) >> 3;
	for (int i = 0; i < len; i ++)
		table [i] = 0;
	return;
} /* BitField::clearall */

bool thesame (const BitField &b1, const BitField &b2, int length)
{
	register int len = (length + 7) >> 3;
	for (int i = 0; i < len; i ++)
		if (b1. table [i] != b2. table [i])
			return 0;
	return 1;
} /* thesame */

unsigned BitField::hashkey (int length) const
{
	int shift = 0;
	unsigned key = 0;
	int maxshift = 27;
	int check32bit = 0xFFFF;
	if (check32bit < 0)
		maxshift = 13;
	int len = (length + 7) >> 3;
	for (int i = 0; i < len; i ++)
	{
		key ^= ((unsigned) table [i]) << shift;
		shift += 7;
		if (shift > maxshift)
			shift -= maxshift;
	}
	return key;
} /* BitField::hashkey */

unsigned BitField::hashadd (int length) const
{
	int shift = 3;
	unsigned key = 0x0F0F;
	int maxshift = 27;
	int check32bit = 0xFFFF;
	if (check32bit < 0)
		maxshift = 13;
	int len = (length + 7) >> 3;
	for (int i = 0; i < len; i ++)
	{
		key ^= ((unsigned) table [i]) << shift;
		shift += 5;
		if (shift > maxshift)
			shift -= maxshift;
	}
	return key;
} /* BitField::hashadd */

int BitField::find (int first, int length) const
{
	// reset the number of the first bit if negative
	if (first < 0)
		first = 0;

	// if the answer is trivial, return it
	if (first >= length)
		return -1;

	// masks for bits with numbers greater than or equal to the given one
	const unsigned char uppermask [] =
		{0xFF, 0xFE, 0xFC, 0xF8, 0xF0, 0xE0, 0xC0, 0x80};
	// masks for bits to consider in the last analyzed byte
	const unsigned char lowermask [] =
		{0xFF, 0x01, 0x03, 0x07, 0x0F, 0x1F, 0x3F, 0x7F};
	// masks for single bits with the given numbers
	const unsigned char bitmask [] =
		{0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80};

	// prepare the numbers of bit masks in the first analyzed byte
	// and in the last analyzed byte
	unsigned char firstmask = uppermask [first & 0x07];
	unsigned char lastmask = lowermask [length & 0x07];

	// figure out the first and last bytes' addresses
	const unsigned char *curbyte = table + (first >> 3);
	const unsigned char *lastbyte = table + ((length - 1) >> 3);

	// if the first byte is the last byte, return the answer
	if (curbyte == lastbyte)
	{
		if (!((*curbyte) & firstmask & lastmask))
			return -1;
	}
	// if there is something in the first byte, return it
	if ((*curbyte) & firstmask)
	{
		int curbit = first & 0x07;
		while (curbit < 8)
			if ((*curbyte) & bitmask [curbit])
				return ((curbyte - table) << 3) + curbit;
			else
				curbit ++;
		throw "Bit not found.";
	}
	// otherwise move to the next byte
	else
		curbyte ++;

	// scan all the bytes but the last one in search for a nonzero bit
	while ((curbyte < lastbyte) && !*curbyte)
		curbyte ++;

	// if found, extract the bit
	if ((curbyte < lastbyte) || (*(lastbyte) & lastmask))
	{
		int curbit = 0;
		while (curbit < 8)
			if ((*curbyte) & bitmask [curbit])
				return ((curbyte - table) << 3) + curbit;
			else
				curbit ++;
		throw "Could not find the bit.";
	}

	return -1;
} /* BitField::find */


// --------------------------------------------------
// ----------------- BITFIELDSET --------------------
// --------------------------------------------------

static int isprime (unsigned n)
// Checks if the number is prime.
// Returns: 1 = Yes, 0 = No.
{
	unsigned i = 2;
	if (n < 2)
		return 0;
	while (i * i <= n)
		if (!(n % i ++))
			return 0;
	return 1;
} /* is prime */

static unsigned primenumber (unsigned n)
// Returns the smallest prime number
// greater than or equal to the given one.
{
	while (!isprime (n))
		n ++;
	return n;
} /* prime number */

// --------------------------------------------------

SetOfBitFields::SetOfBitFields (int len, int maxelem): values ()
{
	if (len <= 0)
		len = 0;
	if (!len)
		maxelem = 0;
	length = len;
	avail = maxelem;
	usedcount = 0;
	if (maxelem)
	{
		size = primenumber (maxelem + (maxelem >> 3));
		bf = new BitField [size];
		buf = new unsigned char [((length + 7) >> 3) * avail];
		if (!bf || !buf)
			throw "Not enough memory for a bit field set.";
		values. allocate (size);
	}
	else
	{
		size = 0;
		bf = NULL;
		buf = NULL;
	}
	bufcur = buf;
	return;
} /* SetOfBitFields::SetOfBitFields */

int SetOfBitFields::add (const BitField &b, int value)
{
	if (!avail)
		return -1;
	unsigned pos = b. hashkey (length) % size;
	unsigned add = 0;
	while (bf [pos]. defined ())
	{
		if (thesame (bf [pos], b, length))
			return values. test (pos);
		if (!add)
		{
			add = b. hashadd (length) % size;
			if (!add)
				add = 1;
		}
		pos += add;
		if (pos >= size)
			pos -= size;
	}
	bf [pos]. define (bufcur, length);
	bufcur += (length + 7) >> 3;
	bf [pos]. takebits (b, length);
	if (value)
		values. set (pos);
	avail --;
	usedcount ++;
	return 2;
} /* SetOfBitFields::add */

int SetOfBitFields::check (const BitField &b) const
{
	if (!size)
		return -1;
	unsigned pos = b. hashkey (length) % size;
	unsigned add = 0;
	while (bf [pos]. defined ())
	{
		if (thesame (bf [pos], b, length))
			return values. test (pos);
		if (!add)
		{
			add = b. hashadd (length) % size;
			if (!add)
				add = 1;
		}
		pos += add;
		if (pos >= size)
			pos -= size;
	}
	return -1;
} /* SetOfBitFields::check */


} // namespace homology
} // namespace chomp

/// @}

